def debug(func):
    def func_wrapper(*args, **kwargs):
        result = func(*args, **kwargs)
        sys.stderr.write(
            'DEBUG: {}({}, {}) = [{}][{}]\n'.format(
                func.__name__,
                args,
                kwargs,
                result,
                result.encode('hex') if isinstance(result, str) else ''
            )
        )
        return result
    return func_wrapper



from .exception import ClientError


class Client(object):
	def __del__(self):
		self.disconnect()

	def __enter__(self):
		return self

	def __exit__(self, exception_type, exception_value, traceback):
		self.disconnect()

	def __init__(self, *args, **kwargs):
		assert 'host' in kwargs, 'No host specified'
		self.snmp_client = None
		self.ssh_client = None

	def can_snmp(self):
		if hasattr(self, 'snmp_client'):
			if self.snmp_client is not None:
				return True
		return False

	def can_ssh(self):
		if hasattr(self, 'ssh_client'):
			if self.ssh_client is not None:
				return True
		return False

	def cli_command(self, command, *args, **kwargs):
		raise NotImplementedError

	def disconnect(self):
		if self.can_snmp():
			self.snmp_client.disconnect()
			del self.snmp_client
		if self.can_ssh():
			if self.ssh_client.is_active():
				self.ssh_client.disconnect()
			del self.ssh_client

	def snmp_get(self, oid):
		if not self.can_snmp():
			raise ClientError('No SNMP client')
		return self.snmp_client.get_oid(self.snmp_client.format_oid(oid))

	def snmp_walk(self, oid):
		if not self.can_snmp():
			raise ClientError('No SNMP client')
		return self.snmp_client.walk_oids([self.snmp_client.format_oid(oid)])

	def software_version(self):
		raise NotImplementedError

	def vendor(self):
		raise NotImplementedError



"""Exceptions for the SSH package"""

class SSH_Exception(Exception):
    """Base exception class"""
    pass


class AuthenticationError(SSH_Exception):
    """Problems with authentication"""
    pass


class ConnectError(SSH_Exception):
    """Client connect errors"""
    pass


class PromptError(SSH_Exception):
    """Prompt related errors"""
    pass


class RecvError(SSH_Exception):
    """Channel receive errors"""
    pass


class ResponseError(SSH_Exception):
    """CLI response errors"""
    pass


class SendError(SSH_Exception):
    """Channel send errors"""
    pass



import sys


class Utils(object):
    @staticmethod
    def debug(func):
        def func_wrapper(*args, **kwargs):
            result = func(*args, **kwargs)
            sys.stderr.write(
                'DEBUG: {}({}, {}) = [{}][{}]\n'.format(
                    func.__name__,
                    args,
                    kwargs,
                    result,
                    result.encode('hex') if isinstance(result, str) else ''
                )
            )
            return result
        return func_wrapper

    @staticmethod
    def group_kwargs(*groups, **kwargs):
        grouped_kwargs = {}
        for key, value in kwargs.iteritems():
            for group in groups:
                group_length = len(group)
                if key.startswith(group):
                    if group not in grouped_kwargs:
                        grouped_kwargs[group] = {}
                    grouped_kwargs[group][key[group_length:]] = value
        return grouped_kwargs



"""Shell class for SSH Client"""

import socket
#
from .exception import PromptError, RecvError, ResponseError, SendError
from .client import Client
from .shell_prompt import ShellPrompt
from .utils import Utils


class Shell(object):
    """SSH shell class"""
    def __del__(self):
        self.exit()

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        self.exit()

    def __init__(self, ssh_client, shell_prompts, optional_args=None):
        assert isinstance(ssh_client, Client)
        assert isinstance(shell_prompts, ShellPrompt)
        if optional_args is None:
            optional_args = {}

        self.ssh_client = ssh_client
        self.shell_prompts = shell_prompts
        self.last_prompt = ''

        self.channel = ssh_client.channel()
        self.channel.settimeout(optional_args.get('timeout', 0.2))
        self.channel.set_combine_stderr(optional_args.get('combine_stderr', False))
        self.channel.get_pty(
            term=optional_args.get('terminal_type', 'dumb'),
            width=optional_args.get('terminal_width', 80),
            height=optional_args.get('terminal_height', 25)
        )
        self.channel.invoke_shell()

        banner, retries_left = self.read_until_prompt()
        if retries_left == 0:
            if len(banner) == 0:
                raise PromptError('Cannot auto-detect prompt')
            timeout_prompt = banner.pop()
            self.shell_prompts.add_prompt(timeout_prompt)
            self.command('\n', 5)
        self.on_banner(banner)

        for command in optional_args.get('initial_commands', []):
            self.command(command)

    def __repr__(self):
        return 'ssh.Shell @ {} ({})'.format(
            hex(long(id(self)) & long(0xffffffff)),
            repr(self.ssh_client)
        )

#    @Utils.debug
    def command(self, command, prompt_retries=40):
        send_command = command.rstrip('\r\n')
        if self.channel.send(send_command + '\r') != (len(send_command) + 1):
            raise SendError('Did not send all of command')
        if prompt_retries:
            while range(10, 0, -1):
                raw_output = self._gets()
                if raw_output != '':
                    output_line = self.on_output_line(raw_output)
                    if output_line != send_command:
                        raise ResponseError('Command echo mismatch')
                    break
            else:
                raise ResponseError('Command not echoed')
        output, retries_left = self.read_until_prompt(prompt_retries)
        if prompt_retries != 0 and retries_left == 0:
            raise PromptError('Prompt not seen')
        return output

    def exit(self, exit_command='exit'):
        if hasattr(self, 'channel') and self.ssh_client.is_active():
            self.channel.send(exit_command)
            self.channel.shutdown(2)
            self.channel.close()
            del self.channel

    def on_banner(self, banner):
        pass

    def on_output_line(self, output_line):
        return output_line.rstrip('\r\n')

    def on_prompt(self, prompt):
        pass

    def read_until_prompt(self, prompt_retries=25):
        output = []
        while prompt_retries:
            raw_output = self._gets()
            if raw_output == '':
                prompt_retries -= 1
            else:
                output_line = self.on_output_line(raw_output)
                if self.shell_prompts.is_prompt(output_line):
                    self.last_prompt = output_line
                    self.on_prompt(output_line)
                    break
                output.append(output_line)
        return (output, prompt_retries)

    @staticmethod
    def _getc(chan):
        if chan.recv_ready():
            return chan.recv(1)
        while True:
            if chan.exit_status_ready():
                break
            try:
                c = chan.recv(1)
            except socket.timeout:
                break
            else:
                if len(c) == 0:
                    raise RecvError('Channel closed during recv()')
                return c
        return None

    def _gets(self):
        s = []
        while True:
            c = self._getc(self.channel)
            if c is None:
                break
            s.append(c)
            if c == '\n':
                break
        return ''.join(s)



import difflib

class Config(object):
	def __init__(self):
		pass

	@staticmethod
	def diff(config_a, config_b, **kwargs):
		from_file = kwargs.get('fromfile', 'a')
		to_file = kwargs.get('tofile', 'b')

		config_diff = []
		for line in  difflib.unified_diff(config_a, config_b, fromfile=from_file, tofile=to_file, n=0, lineterm=''):
			config_diff.append(line)
		return config_diff

	@staticmethod
	def export_concat(export_config):
		export_section = None
		line_buffer = []

		concat_config = []
		for line in export_config:
			if line == '':
				continue
			if line[0] == '/':
				export_section = line
				continue
			elif line[0] == '#':
				continue
			elif line[-1:] == '\\':
				line_buffer.append(line[:-1].lstrip())
			else:
				if export_section is None:
					raise ValueError('No configuration section')
				line_buffer.append(line.lstrip())
				line_buffer.insert(0, '{} '.format(export_section))
				concat_config.append(''.join(line_buffer))
				line_buffer = []
		return concat_config

	@staticmethod
	def reverse(concat_config):
		pass



import sys


class Utils(object):
    @staticmethod
    def debug(func):
        def func_wrapper(*args, **kwargs):
            result = func(*args, **kwargs)
            hex_result = ''
            if isinstance(result, str):
                hex_result = result.encode('hex')
            elif isinstance(result, list):
                if len(result) == 1 and isinstance(result[0], str):
                    hex_result = result[0].encode('hex')
            sys.stderr.write(
                'DEBUG: {}({}, {}) = [{}][{}]\n'.format(
                    func.__name__,
                    args,
                    kwargs,
                    result,
                    hex_result
                )
            )
            return result
        return func_wrapper



from .exception import *
from . import client


class Client(object):
    def __init__(self, ssh_client):
        assert isinstance(ssh_client, client.Client)
        self.ssh_client = ssh_client

    def sftp_client(self):
        return paramiko.SFTPClient.from_transport(self.ssh_client.paramiko_transport)



from pysnmp.entity.rfc3413.oneliner import cmdgen
from pyasn1.type import univ
#
from .exception import *

__all__ = ['Client']

_Client__oid_lookup = {
    'bgp4PathAttrEntry': (1, 3, 6, 1, 2, 1, 15, 6, 1),
    'bgpPeerEntry': (1, 3, 6, 1, 2, 1, 15, 3, 1),
    'dot1dTpFdbEntry': (1, 3, 6, 1, 2, 1, 17, 4, 3, 1),
    'dot1dBasePortEntry': (1, 3, 6, 1, 2, 1, 17, 1, 4, 1),
    'dot1qTpFdbEntry': (1, 3, 6, 1, 2, 1, 17, 7, 1, 2, 2, 1),
    'dot1qVlanCurrentEntry': (1, 3, 6, 1, 2, 1, 17, 7, 1, 4, 2, 1),
    'entPhysicalEntry': (1, 3, 6, 1, 2, 1, 47, 1, 1, 1, 1),
    'entPhysicalMfgName': (1, 3, 6, 1, 2, 1, 47, 1, 1, 1, 1, 12, 1),
    'ifEntry': (1, 3, 6, 1, 2, 1, 2, 2, 1),
    'ifXEntry': (1, 3, 6, 1, 2, 1, 31, 1, 1, 1),
    'inetCidrRouteEntry': (1, 3, 6, 1, 2, 1, 4, 24, 7, 1),
    'ipAddressEntry': (1, 3, 6, 1, 2, 1, 4, 34, 1),
    'ipAddressPrefixEntry': (1, 3, 6, 1, 2, 1, 4, 32, 1),
    'ipNetToMediaEntry': (1, 3, 6, 1, 2, 1, 4, 22, 1),
    'ipRouteEntry': (1, 3, 6, 1, 2, 1, 4, 21, 1),
    'lldpLocPortEntry': (1, 0, 8802, 1, 1, 2, 1, 3, 7, 1),
    'lldpRemEntry': (1, 0, 8802, 1, 1, 2, 1, 4, 1, 1),
    'sysOREntry': (1, 3, 6, 1, 2, 1, 1, 9, 1),
}

class Client(object):
    def __del__(self):
        self.disconnect()

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        self.disconnect()

    def __init__(self, host, community='public', port=161):
        self._transport = cmdgen.UdpTransportTarget((host, port), timeout=1.0, retries=1)
        self._authdata = cmdgen.CommunityData('securityIndex', community)

    def __repr__(self):
        return 'snmp.Client @ {} ({}, {})'.format(
            hex(long(id(self)) & long(0xffffffff)),
            repr(self._transport),
            repr(self._authdata)
        )

    def bgpIdentifier(self):
        return self.get_oid((1, 3, 6, 1, 2, 1, 15, 4, 0))

    def bgpLocalAs(self):
        return self.get_oid((1, 3, 6, 1, 2, 1, 15, 2, 0))

    def bgpPeerTable(self, column_names):
        bgp_peer_table = self._table_entries(
            __oid_lookup['bgpPeerEntry'],
            {
                'bgpPeerIdentifier': 1,
                'bgpPeerState': 2,
                'bgpPeerAdminStatus': 3,
                'bgpPeerNegotiatedVersion': 4,
                'bgpPeerLocalAddr': 5,
                'bgpPeerLocalPort': 6,
                'bgpPeerRemoteAddr': 7,
                'bgpPeerRemotePort': 8,
                'bgpPeerRemoteAs': 9,
                'bgpPeerInUpdates': 10,
                'bgpPeerOutUpdates': 11,
                'bgpPeerInTotalMessages': 12,
                'bgpPeerOutTotalMessages': 13,
                'bgpPeerLastError': 14,
                'bgpPeerFsmEstablishedTransitions': 15,
                'bgpPeerFsmEstablishedTime': 16,
                'bgpPeerConnectRetryInterval': 17,
                'bgpPeerHoldTime': 18,
                'bgpPeerKeepAlive': 19,
                'bgpPeerHoldTimeConfigured': 20,
                'bgpPeerKeepAliveConfigured': 21,
                'bgpPeerMinASOriginationInterval': 22,
                'bgpPeerMinRouteAdvertisementInterval': 23,
                'bgpPeerInUpdateElapsedTime': 24,
            },
            column_names
        )
        for bgp_peer, bgp_peer_entry in bgp_peer_table.iteritems():
            if 'bgpPeerState' in bgp_peer_entry:
                bgp_peer_entry['bgpPeerState'] = self.decode_bgpPeerState(
                    bgp_peer_entry['bgpPeerState']
                )
            if 'bgpPeerAdminStatus' in bgp_peer_entry:
                bgp_peer_entry['bgpPeerAdminStatus'] = self.decode_bgpPeerAdminStatus(
                    bgp_peer_entry['bgpPeerAdminStatus']
                )
        return bgp_peer_table

    def bgp4PathAttrTable(self, column_names):
        return self._table_entries(
            __oid_lookup['bgp4PathAttrEntry'],
            {
                'bgp4PathAttrPeer': 1,
                'bgp4PathAttrIpAddrPrefixLen': 2,
                'bgp4PathAttrIpAddrPrefix': 3,
                'bgp4PathAttrOrigin': 4,
                'bgp4PathAttrASPathSegment': 5,
                'bgp4PathAttrNextHop': 6,
                'bgp4PathAttrMultiExitDisc': 7,
                'bgp4PathAttrLocalPref': 8,
                'bgp4PathAttrAtomicAggregate': 9,
                'bgp4PathAttrAggregatorAS': 10,
                'bgp4PathAttrAggregatorAddr': 11,
                'bgp4PathAttrCalcLocalPref': 12,
                'bgp4PathAttrBest': 13,
                'bgp4PathAttrUnknown': 14,
            },
            column_names
        )

    @staticmethod
    def decode_bgpPeerAdminStatus(decode_value):
        admin_status = {
            1: 'stop',
            2: 'start',
        }
        return admin_status.get(decode_value, '')

    @staticmethod
    def decode_bgpPeerState(decode_value):
        peer_state = {
            1: 'idle',
            2: 'connect',
            3: 'active',
            4: 'opensent',
            5: 'openconfirm',
            6: 'established',
        }
        return peer_state.get(decode_value, '')

    @staticmethod
    def decode_dot1qTpFdbStatus(decode_value):
        dot1qTpFdbStatus = {
            1: 'other',
            2: 'invalid',
            3: 'learned',
            4: 'self',
            5: 'mgmt',
        }
        return dot1qTpFdbStatus.get(decode_value, '')

    @staticmethod
    def decode_ifAdminStatus(decode_value):
        admin_status = {
            1: 'up',
            2: 'down',
            3: 'testing',
        }
        return admin_status.get(decode_value, '')

    @staticmethod
    def decode_ifOperStatus(decode_value):
        oper_status = {
            1: 'up',
            2: 'down',
            3: 'testing',
            4: 'unknown',
            5: 'dormant',
            6: 'notPresent',
            7: 'lowerLayerDown',
        }
        return oper_status.get(decode_value, '')

    @staticmethod
    def decode_PhysicalClass(decode_value):
        physical_class = {
            1: 'other',
            2: 'unknown',
            3: 'chassis',
            4: 'backplane',
            5: 'container',
            6: 'powerSupply',
            7: 'fan',
            8: 'sensor',
            9: 'module',
            10: 'port',
            11: 'stack',
            12: 'cpu',
        }
        return physical_class.get(decode_value, '')

    def dot1dBasePortTable(self, column_names):
        return self._table_entries(
            __oid_lookup['dot1dBasePortEntry'],
            {
                'dot1dBasePort': 1,
                'dot1dBasePortIfIndex': 2,
                'dot1dBasePortCircuit': 3,
                'dot1dBasePortBasePortDelayExceededDiscards': 4,
                'dot1dBasePortBasePortMtuExceededDiscards': 5,
            },
            column_names
        )

    def dot1dTpFdbTable(self, column_names):
        return self._table_entries(
            __oid_lookup['dot1dTpFdbEntry'],
            {
                'dot1dTpFdbAddress': 1,
                'dot1dTpFdbPort': 2,
                'dot1dTpFdbStatus': 3,
            },
            column_names
        )

    def dot1qTpFdbTable(self, column_names):
        dot1q_tp_fdb_table = self._table_entries(
            __oid_lookup['dot1qTpFdbEntry'],
            {
                'dot1qTpFdbAddress': 1,
                'dot1qTpFdbPort': 2,
                'dot1qTpFdbStatus': 3,
            },
            column_names
        )
        for _, dot1q_tp_fdb_entry in dot1q_tp_fdb_table.iteritems():
            if 'dot1qTpFdbStatus' in dot1q_tp_fdb_entry:
                dot1q_tp_fdb_entry['dot1qTpFdbStatus'] = self.decode_dot1qTpFdbStatus(
                    dot1q_tp_fdb_entry.get('dot1qTpFdbStatus')
                )
        return dot1q_tp_fdb_table

    def dot1qVlanCurrentTable(self, column_names):
        return self._table_entries(
            __oid_lookup['dot1qVlanCurrentEntry'],
            {
                'dot1qVlanTimeMark': 1,
                'dot1qVlanIndex': 2,
                'dot1qVlanFdbId': 3,
                'dot1qVlanCurrentEgressPorts': 4,
                'dot1qVlanCurrentUntaggedPorts': 5,
                'dot1qVlanStatus': 6,
                'dot1qVlanCreationTime': 7,
            },
            column_names
        )

    def disconnect(self):
        if hasattr(self, '_authdata'):
            del self._authdata
        if hasattr(self, '_transport'):
            del self._transport

    def enterprise(self):
        sys_object_id = self.sysObjectID()

        vendor = self.get_oid(__oid_lookup['entPhysicalMfgName'])
        if vendor == '':
            if sys_object_id.startswith('1.3.6.1.4.1.2544.1'):
                vendor = 'Adva'
            elif sys_object_id.startswith('1.3.6.1.4.1.30065.1'):
                vendor = 'Arista'
            elif sys_object_id.startswith('1.3.6.1.4.1.9 1'):
                vendor = 'Cisco'
            elif sys_object_id.startswith('1.3.6.1.4.1.40310'):
                vendor = 'Cumulus'
            elif sys_object_id.startswith('1.3.6.1.4.1.1991.1'):
                vendor = 'Foundry'
            elif sys_object_id.startswith('1.3.6.1.4.1.2636.1'):
                vendor = 'Juniper'
            elif sys_object_id.startswith('1.3.6.1.4.1.14988.1'):
                vendor = 'Mikrotik'
            elif sys_object_id.startswith('1.3.6.1.4.1.2352.1'):
                vendor = 'Redback'
            elif sys_object_id.startswith('1.3.6.1.4.1.890.1'):
                vendor = 'Zyxel'
        return vendor, sys_object_id

    def entPhysicalTable(self, column_names):
        return self._table_entries(
            __oid_lookup['entPhysicalEntry'],
            {
                'entPhysicalIndex': 1,
                'entPhysicalDescr': 2,
                'entPhysicalVendorType': 3,
                'entPhysicalContainedIn': 4,
                'entPhysicalClass': 5,
                'entPhysicalParentRelPos': 6,
                'entPhysicalName': 7,
                'entPhysicalHardwareRev': 8,
                'entPhysicalFirmwareRev': 9,
                'entPhysicalSoftwareRev': 10,
                'entPhysicalSerialNum': 11,
                'entPhysicalMfgName': 12,
                'entPhysicalModelName': 13,
                'entPhysicalAlias': 14,
                'entPhysicalAssetID': 15,
                'entPhysicalIsFRU': 16,
                'entPhysicalMfgDate': 17,
                'entPhysicalUris': 18,
            },
            column_names
        )

    @staticmethod
    def format_oid(oid):
        if isinstance(oid, str):
            return tuple(map(int, oid.lstrip('.').split('.')))
        elif isinstance(oid, tuple):
            return oid
        elif isinstance(oid, (int, tuple)):
            raise TypeError(oid)
        else:
            try:
                return tuple(oid)
            except TypeError:
                return (int(oid), )
            raise ValueError('Cannot format OID - [{}]'.format(oid))

    def get_oid(self, base_oid, oid_index=None):
        snmp_oid = base_oid
        if oid_index is not None:
            if isinstance(oid_index, int):
                snmp_oid += (oid_index, )
            elif isinstance(oid_index, tuple):
                snmp_oid += oid_index
            else:
                snmp_oid += self.format_oid(oid_index)

        cmdGen = cmdgen.CommandGenerator()
        cmdGen.lexicographicMode = False
        cmdGen.lookupNames = False
        cmdGen.lookupValues = False

        errorIndication, errorStatus, errorIndex, varBinds = cmdGen.getCmd(
            self._authdata,
            self._transport,
            univ.ObjectIdentifier(snmp_oid),
        )

        if errorIndication:
            raise SNMP_Exception(errorIndication)
        else:
            if errorStatus:
                raise SNMP_Exception(snmp_oid)
            else:
                oid, value = varBinds[0]
                if oid != snmp_oid:
                    raise SNMP_Exception('Mismatched SNMP OID - [{}]'.format(oid))
                return value

        return None

    def get_table_index_only(self, base_oid):
        base_oid_length = len(base_oid)
        return {(k[base_oid_length:], v) for k, v in self.walk_oids([base_oid]).iteritems()}

    def ifIndex_to_ifName(self, if_index):
        if_name = self.get_oid((1, 3, 6, 1, 2, 1, 31, 1, 1, 1, 1), if_index)
        if if_name is None:
            if_name = self.get_oid((1, 3, 6, 1, 2, 1, 2, 2, 1, 1), if_index)
            if if_name is None:
                return if_index
        return if_name

    def ifTable(self, column_names):
        if_table = self._table_entries(
            __oid_lookup['ifEntry'],
            {
                'ifIndex': 1,
                'ifDescr': 2,
                'ifType': 3,
                'ifMtu': 4,
                'ifSpeed': 5,
                'ifPhysAddress': 6,
                'ifAdminStatus': 7,
                'ifOperStatus': 8,
                'ifLastChange': 9,
                'ifInOctets': 10,
                'ifInUcastPkts': 11,
                'ifInNUcastPkts': 12,
                'ifInDiscards': 13,
                'ifInErrors': 14,
                'ifInUnknownProtos': 15,
                'ifOutOctets': 16,
                'ifOutUcastPkts': 17,
                'ifOutNUcastPkts': 18,
                'ifOutDiscards': 19,
                'ifOutErrors': 20,
                'ifOutQLen': 21,
                'ifSpecific': 22,
            },
            column_names
        )
        for if_name, if_entry in if_table.iteritems():
            if 'ifAdminStatus' in if_entry:
                if_entry['ifAdminStatus'] = self.decode_ifAdminStatus(if_entry.get('ifAdminStatus'))
            if 'ifOperStatus' in if_entry:
                if_entry['ifOperStatus'] = self.decode_ifOperStatus(if_entry.get('ifOperStatus'))
        return if_table

    def ifXTable(self, column_names):
        if_table = self._table_entries(
            __oid_lookup['ifXEntry'],
            {
                'ifName': 1,
                'ifInMulticastPkts': 2,
                'ifInBroadcastPkts': 3,
                'ifOutMulticastPkts': 4,
                'ifOutBroadcastPkts': 5,
                'ifHCInOctets': 6,
                'ifHCInUcastPkts': 7,
                'ifHCInMulticastPkts': 8,
                'ifHCInBroadcastPkts': 9,
                'ifHCOutOctets': 10,
                'ifHCOutUcastPkts': 11,
                'ifHCOutMulticastPkts': 12,
                'ifHCOutBroadcastPkts': 13,
                'ifLinkUpDownTrapEnable': 14,
                'ifHighSpeed': 15,
                'ifPromiscuousMode': 16,
                'ifConnectorPresent': 17,
                'ifAlias': 18,
                'ifCounterDiscontinuityTime': 19,
            },
            column_names
        )
        for if_index, if_entry in if_table.iteritems():
            if 'ifLinkUpDownTrapEnable' in if_entry:
                if_entry['ifLinkUpDownTrapEnable'] = if_entry.get('ifLinkUpDownTrapEnable')
            if 'ifConnectorPresent' in if_entry:
                if_entry['ifConnectorPresent'] = if_entry.get('ifConnectorPresent')
        return if_table

    def inetCidrRouteTable(self, column_names):
        return self._table_entries(
            __oid_lookup['inetCidrRouteEntry'],
            {
                'inetCidrRouteDestType': 1,
                'inetCidrRouteDest': 2,
                'inetCidrRoutePfxLen': 3,
                'inetCidrRoutePolicy': 4,
                'inetCidrRouteNextHopType': 5,
                'inetCidrRouteNextHop': 6,
                'inetCidrRouteIfIndex': 7,
                'inetCidrRouteType': 8,
                'inetCidrRouteProto': 9,
                'inetCidrRouteAge': 10,
                'inetCidrRouteNextHopAS': 11,
                'inetCidrRouteMetric1': 12,
                'inetCidrRouteMetric2': 13,
                'inetCidrRouteMetric3': 14,
                'inetCidrRouteMetric4': 15,
                'inetCidrRouteMetric5': 16,
                'inetCidrRouteStatus': 17,
            },
            column_names
        )

    def interfaces(self):
        return self.ifXTable(['ifName'])
#        return self.get_table_index_only((1, 3, 6, 1, 2, 1, 2, 2, 1, 2))

    def ipAddressPrefixTable(self, column_names):
        return self._table_entries(
            __oid_lookup['ipAddressPrefixEntry'],
            {
                'ipAddressPrefixIfIndex': 1,
                'ipAddressPrefixType': 2,
                'ipAddressPrefixPrefix': 3,
                'ipAddressPrefixLength': 4,
                'ipAddressPrefixOrigin': 5,
                'ipAddressPrefixOnLinkFlag': 6,
                'ipAddressPrefixAutonomousFlag': 7,
                'ipAddressPrefixAdvPreferredLifetime': 8,
                'ipAddressPrefixAdvValidLifetime': 9,
            },
            column_names
        )

    def ipAddressTable(self, column_names):
        return self._table_entries(
            __oid_lookup['ipAddressEntry'],
            {
                'ipAddressAddrType': 1,
                'ipAddressAddr': 2,
                'ipAddressIfIndex': 3,
                'ipAddressType': 4,
                'ipAddressPrefix': 5,
                'ipAddressOrigin': 6,
                'ipAddressStatus': 7,
                'ipAddressCreated': 8,
                'ipAddressLastChanged': 9,
                'ipAddressRowStatus': 10,
                'ipAddressStorageType': 11,
            },
            column_names
        )

    def ipNetToMediaTable(self, column_names):
        return self._table_entries(
            __oid_lookup['ipNetToMediaEntry'],
            {
                'ipNetToMediaIfIndex': 1,
                'ipNetToMediaPhysAddress': 2,
                'ipNetToMediaNetAddress': 3,
                'ipNetToMediaMediaType': 4,
            },
            column_names
        )

    def ipRouteTable(self, column_names):
        return self._table_entries(
            __oid_lookup['ipRouteEntry'],
            {
                'ipRouteDest': 1,
                'ipRouteIfIndex': 2,
                'ipRouteMetric1': 3,
                'ipRouteMetric2': 4,
                'ipRouteMetric3': 5,
                'ipRouteMetric4': 6,
                'ipRouteNextHop': 7,
                'ipRouteType': 8,
                'ipRouteProto': 9,
                'ipRouteAge': 10,
                'ipRouteMask': 11,
                'ipRouteMetric5': 12,
                'ipRouteInfo': 13,
            },
            column_names
        )

    def lldpLocPortTable(self, column_names):
        return self._table_entries(
            __oid_lookup['lldpLocPortEntry'],
            {
                'lldpLocPortNum': 1,
                'lldpLocPortIdSubtype': 2,
                'lldpLocPortId': 3,
                'lldpLocPortDesc': 4,
            },
            column_names
        )

    def lldpRemTable(self, column_names):
        return self._table_entries(
            __oid_lookup['lldpRemEntry'],
            {
                'lldpRemTimeMark': 1,
                'lldpRemLocalPortNum': 2,
                'lldpRemIndex': 3,
                'lldpRemChassisIdSubtype': 4,
                'lldpRemChassisId': 5,
                'lldpRemPortIdSubtype': 6,
                'lldpRemPortId': 7,
                'lldpRemPortDesc': 8,
                'lldpRemSysName': 9,
                'lldpRemSysDesc': 10,
                'lldpRemSysCapSupported': 11,
                'lldpRemSysCapEnabled': 12,
            },
            column_names
        )

    def set_oid(self, oid):
        raise NotImplementedError

    def sysDescr(self):
        return str(self.get_oid((1, 3, 6, 1, 2, 1, 1, 1, 0)))

    def sysObjectID(self):
        sys_object_id = '.'.join(map(str, self.get_oid((1, 3, 6, 1, 2, 1, 1, 2, 0))))
        if len(sys_object_id) == 0:
            raise ValueError('No sysObjectID')
        return sys_object_id

    def sysORLastChange(self):
        return str(self.get_oid((1, 3, 6, 1, 2, 1, 1, 8, 0)))

    def sysORTable(self, column_names):
        return self._table_entries(
            __oid_lookup['sysOREntry'],
            {
                'sysORIndex': 1,
                'sysORID': 2,
                'sysORDescr': 3,
                'sysORUpTime': 4,
            },
            column_names
        )

    def sysUpTime(self):
        return int(self.get_oid((1, 3, 6, 1, 2, 1, 1, 3, 0)))

    def sysContact(self):
        return str(self.get_oid((1, 3, 6, 1, 2, 1, 1, 4, 0)))

    def sysName(self):
        return str(self.get_oid((1, 3, 6, 1, 2, 1, 1, 5, 0)))

    def sysLocation(self):
        return str(self.get_oid((1, 3, 6, 1, 2, 1, 1, 6, 0)))

    def sysServices(self):
        return self.get_oid((1, 3, 6, 1, 2, 1, 1, 7, 0))

    def walk_oids(self, oids):
        assert isinstance(oids, list)

        cmdGen = cmdgen.CommandGenerator()
        cmdGen.lexicographicMode = False
        cmdGen.lookupNames = False
        cmdGen.lookupValues = False

        errorIndication, errorStatus, errorIndex, varBindTable = cmdGen.nextCmd(
            self._authdata,
            self._transport,
            *map(univ.ObjectIdentifier, oids)
        )

        if errorIndication:
            raise SNMP_Exception(errorIndication)
        else:
            if errorStatus:
                raise SNMP_Exception(errorStatus)
            else:
                ret_ = {}
                for varBindTableRow in varBindTable:
                    for oid, value in varBindTableRow:
                        if oid in ret_:
                            if ret_[oid] == value:
                                continue
                            raise SNMP_Exception('OID already seen - [{}]'.format(oid))
                        ret_[oid] = value
                return ret_
        return None

    def _table_entries(self, entry_oid, table_columns, column_names):
        assert isinstance(table_columns, dict)
        assert isinstance(column_names, list)

        if column_names[0] == '*':
            column_names = table_columns.keys()

        oid_list = []
        for column_name in column_names:
            if column_name in table_columns:
#                oid_list.append(entry_oid + (table_columns[column_name], ))
                pass
            else:
                raise ValueError('Invalid column - [{}]'.format(column_name))

        if len(column_names) == 1:
            oid_list.append(entry_oid + (table_columns[column_names[0]], ))
        else:
            oid_list.append(entry_oid)

        entry_oid_length = len(entry_oid)
        table_columns_reversed = dict((v, k) for k, v in table_columns.iteritems())

        table_entries = {}
        for oid, value in self.walk_oids(oid_list).iteritems():
            if oid[:entry_oid_length] != entry_oid:
                raise ValueError('OID out of range - [{}]'.format(oid))

            oid_parts = oid[entry_oid_length:]
            if len(oid_parts) < 2:
                raise ValueError('Cannot get index')

            table_entry = oid_parts[0]
            if table_entry not in table_columns_reversed:
                raise SNMP_Exception('No column name for {}'.format(table_entry))
            table_entry = table_columns_reversed[table_entry]

            table_index = tuple(oid_parts[1:])
            if table_index not in table_entries:
                table_entries[table_index] = {}

            table_entries[table_index][table_entry] = value
        return table_entries



import re
#
from yandc_base.exception import ClientError


class Utils(object):
    @staticmethod
    def export_concat(export_config):
        export_section = None
        line_buffer = []

        concat_config = []
        for line in export_config:
            if line == '':
                continue
            if line[0] == '/':
                export_section = line
                continue
            elif line[0] == '#':
                continue
            elif line[-1:] == '\\':
                line_buffer.append(line[:-1].lstrip())
            else:
                if export_section is None:
                    raise ValueError('No configuration section')
                line_buffer.append(line.lstrip())
                line_buffer.insert(0, '{} '.format(export_section))
                concat_config.append(''.join(line_buffer))
                line_buffer = []
        return concat_config

    @staticmethod
    def index_values(values, key='name'):
        values_indexed = {}
        for value in values:
            if key in value:
                if value[key] not in values_indexed:
                    values_indexed[value[key]] = []
                values_indexed[value[key]].append(value)
            else:
                raise KeyError('Key not seen - [{}]'.format(key))
        return values_indexed

    @staticmethod
    def is_cli_error(output_line):
        re_match = re.match(r'[^\(]+\(line \d+ column \d+\)', output_line)
        if re_match is not None:
            return True
        return False

    @staticmethod
    def is_cli_warning(output_line):
        re_match = re.match(r'\[([^\]]+)\]$', output_line)
        if re_match is not None:
            return True
        return False

    @staticmethod
    def parse_as_key_value(keys_and_values):
        as_key_value = {}

        last_key = None
        for kv_part in keys_and_values:
            if kv_part.find('=') != -1:
                key, value = kv_part.split('=', 1)
                if key in as_key_value:
                    raise KeyError('Key already seen - [{}]'.format(key))
                as_key_value[key] = value
                last_key = key
            elif last_key is not None:
                as_key_value[last_key] += ' {}'.format(kv_part)
            else:
                raise ClientError(kv_part)
        return as_key_value

    @classmethod
    def parse_print_as_value(cls, print_as_values):
        as_value = []
        for line in print_as_values.replace('.id=*', '\n.id=*').splitlines():
            if line == '':
                continue
            as_value.append(cls.parse_as_key_value(line.split(';')))
        return as_value

    @staticmethod
    def print_concat(print_output):
        concat = []
        for line in print_output:
            if line == '':
                continue
            if line.startswith('   '):
                concat[-1] = ' '.join([concat[-1], line.strip()])
            else:
                if line.find(';;; ') != -1:
                    line = line.replace(';;; ', 'comment=')
                concat.append(line.strip())
        return concat

    @staticmethod
    def print_to_values(print_output):
        to_values = {}
        for line in print_output:
            if line == '':
                continue
            key, value = line.split(':', 1)
            if key in to_values:
                raise KeyError('Key already seen - [{}]'.format(key))
            to_values[key.strip()] = value.strip()
        return to_values

    @classmethod
    def print_to_values_structured(cls, print_output):
        to_values_structured = []
        for line in print_output:
            line_parts = line.strip().split()
            if len(line_parts) == 0:
                continue
            if line_parts[0].isdigit():
                index_seen = line_parts.pop(0)
            else:
                index_seen = -1
            flags_seen = ''
            while True:
                if not len(line_parts):
                    raise ValueError('No parts available')
                part = line_parts.pop(0)
                if part.find('=') == -1:
                    flags_seen += part
                else:
                    line_parts.insert(0, part)
                    line_parts.insert(0, 'flags={}'.format(flags_seen))
                    line_parts.insert(0, 'index={}'.format(index_seen))
                    break
            to_values_structured.append(cls.parse_as_key_value(line_parts))
        return to_values_structured

    @staticmethod
    def to_seconds(time_format):
        seconds = minutes = hours = days = weeks = 0

        number_buffer = ''
        for current_character in time_format:
            if current_character.isdigit():
                number_buffer += current_character
                continue
            if current_character == 's':
                seconds = int(number_buffer)
            elif current_character == 'm':
                minutes = int(number_buffer)
            elif current_character == 'h':
                hours = int(number_buffer)
            elif current_character == 'd':
                days = int(number_buffer)
            elif current_character == 'w':
                weeks = int(number_buffer)
            else:
                raise ValueError('Invalid specifier - [{}]'.format(current_character))
            number_buffer = ''

        seconds += (minutes * 60)
        seconds += (hours * 3600)
        seconds += (days * 86400)
        seconds += (weeks * 604800)

        return seconds



from yandc_cumulus import Client as CL_Client
from yandc_eos import Client as EOS_Client
from yandc_ios import Client as IOS_Client
from yandc_iosxr import Client as XR_Client
from yandc_ros import Client as ROS_Client


class ConnectHandler(object):
    def __del__(self):
        self.disconnect()

    def __init__(self, *args, **kwargs):
        client_kwargs = {}

        if 'host' in kwargs:
            client_kwargs['host'] = kwargs['host']
        elif 'ip' in kwargs:
            client_kwargs['host'] = kwargs['ip']
        else:
            raise ValueError('No Host or IP specified')

        if 'username' in kwargs:
            client_kwargs['username'] = kwargs['username']
        if 'password' in kwargs:
            client_kwargs['password'] = kwargs['password']
        if 'port' in kwargs:
            client_kwargs['ssh_port'] = kwargs['port']
        print client_kwargs

        if kwargs['device_type'] == 'arista':
            self.yandc_client = EOS_Client(**client_kwargs)
        elif kwargs['device_type'] == 'cisco_ios':
            self.yandc_client = IOS_Client(**client_kwargs)
        elif kwargs['device_type'] == 'cisco_xr':
            self.yandc_client = XR_Client(**client_kwargs)
        elif kwargs['device_type'] == 'cumulus_linux':
            self.yandc_client = CL_Client(**client_kwargs)
        elif kwargs['device_type'] == 'mikrotik':
            self.yandc_client = ROS_Client(**client_kwargs)
        else:
            raise ValueError('Device type ' + kwargs['device_type'] + ' not supported')

    def check_config_mode(self):
        return self.yandc_client.in_configure_mode()

    def config_mode(self, config_command='', *args):
        return NotImplementedError

    def disconnect(self):
        if hasattr(self, 'yandc_client'):
            self.yandc_client.disconnect()
            del self.yandc_client

    def enable(self):
        return ''

    def exit_config_mode(self, exit_config, *args):
        return NotImplementedError

    def find_prompt(self, *args):
        return self.yandc_client.ssh_shell.last_prompt

    def send_command(self, command, *args):
        return self.yandc_client.cli_command(command)

    def send_config_from_file(self, config_file, **kwargs):
        with open(config_file, 'rU') as f:
            self.send_config_set(f.read().splitlines())

    def send_config_set(self, config_commands, *args):
        return self.yandc_client.configure_via_cli(config_commands)



class ClientException(Exception):
    pass


class ClientError(ClientException):
    pass


class DeviceMismatchError(ClientException):
    """Device doesn't match SNMP"""
    pass


class FileError(ClientException):
    """File related errors"""
    pass



import datetime
import StringIO
#
import yandc_base as base
try:
	from .snmp import Client as SNMP_Client
except ImportError:
	HAVE_SNMP = False
else:
	from yandc_snmp.exception import SNMP_Exception
	HAVE_SNMP = True
import yandc_ssh as ssh
#
from .shell import Shell
from .utils import Utils

try:
	import rosapi
except ImportError:
	HAVE_APILIB = False
else:
	HAVE_APILIB = True


class Client(base.Client):
	def __init__(self, *args, **kwargs):
		super(Client, self).__init__(*args, **kwargs)

		grouped_kwargs = base.Utils.group_kwargs('snmp_', 'ssh_', 'rosapi_', **kwargs)

		if HAVE_SNMP and 'snmp_' in grouped_kwargs:
			snmp_client = SNMP_Client(kwargs['host'], **grouped_kwargs['snmp_'])
			try:
				sys_object_id = snmp_client.sysObjectID()
			except SNMP_Exception:
				pass
			else:
				if not self.is_mikrotik(sys_object_id):
					raise base.ClientError('Not a Mikrotik device')
				self.snmp_client = snmp_client

		self.in_safe_mode = False
		self._cli_output_cache = {}

		if HAVE_APILIB and 'rosapi_' in grouped_kwargs:
			pass

		if not self.can_rosapi():
			if 'ssh_' not in grouped_kwargs:
				grouped_kwargs['ssh_'] = {}
			if 'username' not in grouped_kwargs['ssh_'] and 'username' in kwargs:
				grouped_kwargs['ssh_']['username'] = kwargs['username']
			if 'password' not in grouped_kwargs['ssh_'] and 'password' in kwargs:
				grouped_kwargs['ssh_']['password'] = kwargs['password']

			shell_prompts = ssh.ShellPrompt(
				ssh.ShellPrompt.regexp_prompt(r'\[[^\@]+\@[^\]]+\] > $')
			)
			shell_prompts.add_prompt(
				ssh.ShellPrompt.regexp_prompt(r'\[[^\@]+\@[^\]]+\] <SAFE> $')
			)

			if 'username' in grouped_kwargs['ssh_']:
				original_username = grouped_kwargs['ssh_']['username']
				grouped_kwargs['ssh_']['username'] += '+ct0h160w'

				if self.can_snmp():
					shell_prompts.add_prompt(
						'[' + original_username + '@' + self.snmp_client.sysName() + '] > '
					)
					shell_prompts.add_prompt(
						'[' + original_username + '@' + self.snmp_client.sysName() + '] <SAFE> '
					)

			self.ssh_client = ssh.Client(kwargs['host'], **grouped_kwargs['ssh_'])

			shell_args = {
				'combine_stderr': True,
			}
			self.ssh_shell = Shell(self.ssh_client, shell_prompts, optional_args=shell_args)

			self._datetime_offset = datetime.datetime.now() - self.ros_datetime()

	def __repr__(self):
		return 'yandc_ros.Client @ {} ({})'.format(
			hex(long(id(self)) & long(0xffffffff)),
			repr(self.ssh_client)
		)

	def cli_command(self, command, *args, **kwargs):
		"""Run the specified command using the Routerboard CLI"""
		use_cache = kwargs.pop('use_cache', False)
		if use_cache:
			if command in self._cli_output_cache:
				return self._cli_output_cache[command]
		else:
			self._cli_output_cache.pop(command, None)
		ssh_output = self.ssh_command(command, *args, **kwargs)
		if use_cache:
			self._cli_output_cache[command] = ssh_output
		return ssh_output
#		raise base.ClientError('No CLI handler')

	def can_rosapi(self):
		"""Can make use of the ROSAPI package"""
		if hasattr(self, '_rosapi'):
			return True
		return False

	def configure_via_cli(self, config_commands):
		if not self.in_safe_mode:
			self.safe_mode_toggle()
		for config_line in config_commands:
			cli_output = self.cli_command(config_line, use_cache=False)
			if cli_output != []:
				raise base.ClientError(cli_output[0])
		self.safe_mode_toggle()

	def disconnect(self):
		if self.can_rosapi():
			pass
		if self.can_ssh() and hasattr(self, 'ssh_shell'):
			if self.in_safe_mode:
				self.safe_mode_toggle()
			self.ssh_shell.exit('/quit')
			del self.ssh_shell
		super(Client, self).disconnect()

	def get_config(self, source=None, section=None):
		if section is not None:
			return self.cli_command('/' + section + ' export verbose', use_cache=False)
		return self.cli_command('/export verbose', use_cache=False)

	def get_interface_config(self, if_name):
		if_type = self.interface_type(if_name)
		if if_type == 'ether':
			if_type = 'ethernet'
		elif if_type == 'pppoe-out':
			if_type = 'pppoe-client'
		return Utils.index_values(
			Utils.print_to_values_structured(
				self.cli_command('/interface {} print without-paging terse'.format(if_type))
			)
		).get(if_name)

	def interface_type(self, if_name):
		indexed_values = Utils.index_values(
			Utils.print_to_values_structured(
				self.cli_command('/interface print without-paging terse', use_cache=True)
			)
		)
		if if_name in indexed_values:
			return indexed_values[if_name][0].get('type')
		return None

	def Xinterfaces(self):
		if self.can_snmp():
			return [str(value.get('ifName')) for value in self.snmp_client.interfaces().values()]
		return [value.get('name') for value in Utils.print_to_values_structured(
			self.cli_command('/interface print without-paging terse')
		)]

	def is_directory(self, candidate_directory):
		indexed_values = Utils.index_values(
			Utils.print_to_values_structured(
				self.cli_command('/file print without-paging terse')
			)
		)
		if candidate_directory in indexed_values:
			if indexed_values[candidate_directory]['type'] == 'directory':
				return True
		return False

	def is_file(self, candidate_file):
		indexed_values = Utils.index_values(
			Utils.print_to_values_structured(
				self.cli_command('/file print without-paging terse')
			)
		)
		if candidate_file in indexed_values:
			if indexed_values[candidate_file]['type'] != 'directory':
				return True
		return False

	@staticmethod
	def is_mikrotik(sys_object_id):
		if sys_object_id.startswith('1.3.6.1.4.1.14988.1'):
			return True
		return False

	def is_slave_interface(self, if_name):
		return Utils.index_values(
			Utils.print_to_values_structured(
				self.cli_command('/interface print without-paging terse', use_cache=True)
			)
		).get(if_name, [])[0].get('flags', '').find('S') == -1

	def master_port(self, if_name):
		indexed_values = Utils.index_values(
			Utils.print_to_values_structured(
				self.cli_command('/interface ethernet print without-paging terse', use_cache=True)
			)
		)
		if if_name in indexed_values:
			master_port = indexed_values[if_name][0].get('master-port')
			if master_port is not None and master_port != 'none':
				return master_port
		return None

	def ros_datetime(self):
		system_clock = Utils.print_to_values(
			self.cli_command('/system clock print without-paging', use_cache=False)
		)
		date_string = '{} {} {}'.format(system_clock['date'], system_clock['time'], 'GMT')
		return datetime.datetime.strptime(date_string, '%b/%d/%Y %H:%M:%S %Z')

	def safe_mode_exit(self):
		if self.ssh_shell.channel.send(chr(0x04)) != 1:
			raise base.ClientError('Failed to send ^X')
		self.in_safe_mode = False

	def safe_mode_toggle(self):
		if self.ssh_shell.channel.send(chr(0x18)) != 1:
			raise base.ClientError('Failed to send ^D')
		hijack_safe_mode = \
			"Hijacking Safe Mode from someone - unroll/release/don't take it [u/r/d]: "
		self.ssh_shell.shell_prompts.add_prompt(hijack_safe_mode)
		shell_output, _ = self.ssh_shell.read_until_prompt(10)
		if self.ssh_shell.last_prompt == hijack_safe_mode:
			self.ssh_shell.channel.send('r')
			shell_output, _ = self.ssh_shell.read_until_prompt(10)
			first_line = shell_output.pop(0)
			if first_line == '114':
				pass
		if len(shell_output) < 2:
			raise base.ClientError(shell_output[0])
		if shell_output[1] == '[Safe Mode taken]':
			if self.in_safe_mode:
				raise base.ClientError('Mismatch with in_safe_mode')
			self.in_safe_mode = True
		elif shell_output[1] == '[Safe Mode released]':
			if not self.in_safe_mode:
				raise base.ClientError('Mismatch with in_safe_mode')
			self.in_safe_mode = False
		else:
			raise base.ClientError(shell_output[1])
		return self.in_safe_mode

	def software_version(self):
		if self.can_snmp():
			return self.snmp_client.os_version()
		elif self.can_ssh():
			return Utils.print_to_values(
				self.ssh_command('/system resource print without-paging')
			).get('version', '')
		return ''

	def ssh_command(self, *args, **kwargs):
		if not self.can_ssh():
			raise base.ClientError('No SSH client')
		if not hasattr(self, 'ssh_shell'):
			raise base.ClientError('No shell channel')
		shell_output = self.ssh_shell.command(*args, **kwargs)
		if self.ssh_shell.last_prompt.endswith(' <SAFE> '):
			if not self.in_safe_mode:
				raise base.ClientError('Mismatch between in_safe_mode and prompt')
		return shell_output

	def system_package_enabled(self, package):
		return Utils.index_values(
			Utils.print_to_values_structured(
				self.cli_command('/system package print without-paging terse', use_cache=True)
			)
		).get(package, [])[0].get('flags', '').find('X') == -1

	def to_seconds_date_time(self, date_time):
		try:
			time_then = datetime.datetime.strptime(date_time, '%b/%d/%Y %H:%M:%S')
		except ValueError:
			return -1
		time_diff = datetime.datetime.now() + self._datetime_offset - time_then
		return int(time_diff.total_seconds())

	def upload_config(self, config, config_name):
		sftp_client = self.ssh_client.sftp_client()
		config_file = StringIO.StringIO('\r\n'.join(config))
		sftp_client.putfo(config_file, '{}.rsc'.format(config_name))
		sftp_client.close()

	@staticmethod
	def vendor():
		return ('Mikrotik', 'ROS')

	@staticmethod
	def write_file_size_check(contents):
		if len(contents) > 4095:
			raise base.FileError('Maximum file size exceeded')

	def write_file_contents(self, name, contents=''):
		self.write_file_size_check(contents)
		cli_output = self.cli_command(
			'/file set {} contents="{}"'.format(name, contents),
			use_cache=False
		)
		if cli_output != []:
			raise base.FileError('Cannot set contents - [{}]'.format(cli_output[0]))
		return True

	def write_rsc_file(self, name, contents=''):
		self.write_file_size_check(contents)
		cli_output = self.cli_command(
			'/system routerboard export file={}'.format(name),
			use_cache=False
		)
		if cli_output != []:
			raise base.FileError('Cannot create file - [{}]'.format(cli_output[0]))
		return self.write_file_contents('{}.rsc'.format(name), contents)

	def write_txt_file(self, name, contents=''):
		self.write_file_size_check(contents)
		cli_output = self.cli_command('/file print file={}'.format(name), use_cache=False)
		if cli_output != []:
			raise base.FileError('Cannot create file - [{}]'.format(cli_output[0]))
		return self.write_file_contents('{}.txt'.format(name), contents)



"""Shell prompt class for SSH Client"""

import re
#
from .exception import PromptError


class ShellPrompt(object):
    """Prompt handling class for Shell"""
    def __init__(self, prompt=None):
        self.prompts = {}
        if prompt is not None:
            self.add_prompt(prompt)

    def __repr__(self):
        return '{}.prompts={}'.format(type(self).__name__, repr(self.prompts))

    def add_prompt(self, prompt):
        if isinstance(prompt, basestring):
            if prompt not in self.prompts:
                self.prompts[prompt] = {
                    'prompt_type': basestring,
                    'prompt_value': prompt
                }
        elif isinstance(prompt, dict):
            if 'prompt_type' in prompt and 'prompt_value' in prompt:
                if prompt['prompt_value'] not in self.prompts:
                    self.prompts[prompt['prompt_value']] = prompt
            else:
                raise PromptError('Invalid prompt specified')
        else:
            raise PromptError('Unsupported prompt type - [{}]'.format(type(prompt)))

    def is_prompt(self, candidate_prompt):
        if candidate_prompt in self.prompts and \
                self.prompts[candidate_prompt]['prompt_type'] is basestring:
            return True
        for prompt in self.prompts.values():
            if prompt['prompt_type'] is basestring:
                continue
            elif prompt['prompt_type'] == 'regexp':
                if re.match(prompt['prompt_regexp'], candidate_prompt):
                    return True
            else:
                raise PromptError(
                    'Unsupported prompt type - [{}]'.format(prompt['prompt_type'])
                )
        return False

    @staticmethod
    def regexp_prompt(re_prompt):
        return {
            'prompt_type': 'regexp',
            'prompt_value': re_prompt,
            'prompt_regexp': re.compile(re_prompt)
        }



from yandc_snmp import Client as BaseClient


class Client(BaseClient):
    def __init__(self, *args, **kwargs):
        self.oid_lookup = {
            'mtXRouterOs': (1, 3, 6, 1, 4, 1, 14988, 1, 1),
            'mtxrWireless': (1, 3, 6, 1, 4, 1, 14988, 1, 1, 1),
            'mtxrWlStatEntry': (1, 3, 6, 1, 4, 1, 14988, 1, 1, 1, 1, 1),
            'mtxrWlRtabEntry': (1, 3, 6, 1, 4, 1, 14988, 1, 1, 1, 2, 1),
            'mtxrWlApEntry': (1, 3, 6, 1, 4, 1, 14988, 1, 1, 1, 3, 1),
            'mtxrQueueSimpleEntry': (1, 3, 6, 1, 4, 1, 14988, 1, 1, 2, 1, 1),
            'mtxrHealth': (1, 3, 6, 1, 4, 1, 14988, 1, 1, 3),
            'mtrxLicVersion': (1, 3, 6, 1, 4, 1, 14988, 1, 1, 4, 4, 0),
            'mtxrSystem': (1, 3, 6, 1, 4, 1, 14988, 1, 1, 7),
            'mtxrFirmwareVersion': (1, 3, 6, 1, 4, 1, 14988, 1, 1, 7, 4, 0),
            'mtxrNeighborTableEntry': (1, 3, 6, 1, 4, 1, 14988, 1, 1, 11, 1, 1),
            'mtxrInterfaceStatsEntry': (1, 3, 6, 1, 4, 1, 14988, 1, 1, 14, 1, 1),
            'mtxrPartitionEntry': (1, 3, 6, 1, 4, 1, 14988, 1, 1, 17, 1, 1),
        }
        super(Client, self).__init__(*args, **kwargs)

    def fw_version(self):
        return self.get_oid(self.oid_lookup['mtxrFirmwareVersion'])

    def mtxr_health(self, column_names):
        table_columns = {
            'mtxrHlCoreVoltage': 1,
            'mtxrHlThreeDotThreeVoltage': 2,
            'mtxrHlFiveVoltage': 3,
            'mtxrHlTwelveVoltage': 4,
            'mtxrHlSensorTemperature': 5,
            'mtxrHlCpuTemperature': 6,
            'mtxrHlBoardTemperature': 7,
            'mtxrHlVoltage': 8,
            'mtxrHlActiveFan': 9,
            'mtxrHlTemperature': 10,
            'mtxrHlProcessorTemperature': 11,
            'mtxrHlPower': 12,
            'mtxrHlCurrent': 13,
            'mtxrHlProcessorFrequency': 14,
            'mtxrHlPowerSupplyState': 15,
            'mtxrHlBackupPowerSupplyState': 16,
            'mtxrHlFanSpeed1': 17,
            'mtxrHlFanSpeed2': 18,
        }

        health = self._table_entries(
            self.oid_lookup['mtxrHealth'],
            table_columns,
            column_names
        )[(0, )]

        if health.get('mtxrHlActiveFan', '') == 'n/a':
            del health['mtxrHlActiveFan']
        return health

    def mtxr_system(self, column_names):
        table_columns = {
            'mtxrSystemReboot': 1,
            'mtxrSystemUSBPowerReset': 2,
            'mtxrSystemSerialNumber': 3,
            'mtxrSystemFirmwareVersion': 4,
            'mtxrSystemNote': 5,
            'mtxrSystemBuildTime': 6,
            'mtxrSystemFirmwareUpgradeVersion': 7,
        }
        return self._table_entries(
            self.oid_lookup['mtxrSystem'],
            table_columns,
            column_names
        )[(0, )]

    def mtxrInterfaceStatsTable(self, column_names):
        table_columns = {
            'mtxrInterfaceStatsIndex': 1,
            'mtxrInterfaceStatsName': 2,
            'mtxrInterfaceStatsDriverRxBytes': 11,
            'mtxrInterfaceStatsDriverRxPackets': 12,
            'mtxrInterfaceStatsDriverTxBytes': 13,
            'mtxrInterfaceStatsDriverTxPackets': 14,
            'mtxrInterfaceStatsTxRx64': 15,
            'mtxrInterfaceStatsTxRx65to127': 16,
            'mtxrInterfaceStatsTxRx128to255': 17,
            'mtxrInterfaceStatsTxRx256to511': 18,
            'mtxrInterfaceStatsTxRx512to1023': 19,
            'mtxrInterfaceStatsTxRx1024to1518': 20,
            'mtxrInterfaceStatsTxRx1519toMax': 21,
            'mtxrInterfaceStatsRxBytes': 31,
            'mtxrInterfaceStatsRxPackets': 32,
            'mtxrInterfaceStatsRxTooShort': 33,
            'mtxrInterfaceStatsRx64': 34,
            'mtxrInterfaceStatsRx64to127': 35,
            'mtxrInterfaceStatsRx128to255': 36,
            'mtxrInterfaceStatsRx256to511': 37,
            'mtxrInterfaceStatsRx512to1023': 38,
            'mtxrInterfaceStatsRx1024to1518': 39,
            'mtxrInterfaceStatsRx1519toMax': 40,
            'mtxrInterfaceStatsRxTooLong': 41,
            'mtxrInterfaceStatsRxBroadcast': 42,
            'mtxrInterfaceStatsRxPause': 43,
            'mtxrInterfaceStatsRxMulticast': 44,
            'mtxrInterfaceStatsRxFCSError': 45,
            'mtxrInterfaceStatsRxAlignError': 46,
            'mtxrInterfaceStatsRxFragement': 47,
            'mtxrInterfaceStatsRxOverflow': 48,
            'mtxrInterfaceStatsRxControl': 49,
            'mtxrInterfaceStatsRxUnknownOp': 50,
            'mtxrInterfaceStatsRxLengthError': 51,
            'mtxrInterfaceStatsRxCodeError': 52,
            'mtxrInterfaceStatsRxCarrierError': 53,
            'mtxrInterfaceStatsRxJabber': 54,
            'mtxrInterfaceStatsRxDrop': 55,
            'mtxrInterfaceStatsTxBytes': 61,
            'mtxrInterfaceStatsTxPackets': 62,
            'mtxrInterfaceStatsTxTooShort': 63,
            'mtxrInterfaceStatsTx64': 64,
            'mtxrInterfaceStatsTx64to127': 65,
            'mtxrInterfaceStatsTx128to255': 66,
            'mtxrInterfaceStatsTx256to511': 67,
            'mtxrInterfaceStatsTx512to1023': 68,
            'mtxrInterfaceStatsTx1024to1518': 69,
            'mtxrInterfaceStatsTx1519toMax': 70,
            'mtxrInterfaceStatsTxTooLong': 71,
            'mtxrInterfaceStatsTxBroadcast': 72,
            'mtxrInterfaceStatsTxPause': 73,
            'mtxrInterfaceStatsTxMulticast': 74,
            'mtxrInterfaceStatsTxUnderrun': 75,
            'mtxrInterfaceStatsTxCollision': 76,
            'mtxrInterfaceStatsTxExcessiveCollision': 77,
            'mtxrInterfaceStatsTxMultipleCollision': 78,
            'mtxrInterfaceStatsTxSingleCollision': 79,
            'mtxrInterfaceStatsTxExcessiveDeferred': 80,
            'mtxrInterfaceStatsTxDeferred': 81,
            'mtxrInterfaceStatsTxLateCollision': 82,
            'mtxrInterfaceStatsTxTotalCollision': 83,
            'mtxrInterfaceStatsTxPauseHonored': 84,
            'mtxrInterfaceStatsTxDrop': 85,
            'mtxrInterfaceStatsTxJabber': 86,
            'mtxrInterfaceStatsTxFCSError': 87,
            'mtxrInterfaceStatsTxControl': 88,
            'mtxrInterfaceStatsTxFragment': 89,
        }
        return self._table_entries(
            self.oid_lookup['mtxrInterfaceStatsEntry'],
            table_columns,
            column_names
        )

    def mtxrNeighborTable(self, column_names):
        table_columns = {
            'mtxrNeighborIndex': 1,
            'mtxrNeighborIpAddress': 2,
            'mtxrNeighborMacAddress': 3,
            'mtxrNeighborVersion': 4,
            'mtxrNeighborPlatform': 5,
            'mtxrNeighborIdentity': 6,
            'mtxrNeighborSoftwareID': 7,
            'mtxrNeighborInterfaceID': 8,
        }
        return self._table_entries(
            self.oid_lookup['mtxrNeighborTableEntry'],
            table_columns,
            column_names
        )

    def mtxrPartitionTable(self, column_names):
        table_columns = {
            'mtxrPartitionIndex': 1,
            'mtxrPartitionName': 2,
            'mtxrPartitionSize': 3,
            'mtxrPartitionVersion': 4,
            'mtxrPartitionActive': 5,
            'mtxrPartitionRunning': 6,
        }
        return self._table_entries(
            self.oid_lookup['mtxrPartitionEntry'],
            table_columns,
            column_names
        )

    def mtxrQueueSimpleTable(self, column_names):
        table_columns = {
            'mtxrQueueSimpleIndex': 1,
            'mtxrQueueSimpleName': 2,
            'mtxrQueueSimpleSrcAddr': 3,
            'mtxrQueueSimpleSrcMask': 4,
            'mtxrQueueSimpleDstAddr': 5,
            'mtxrQueueSimpleDstMask': 6,
            'mtxrQueueSimpleIface': 7,
            'mtxrQueueSimpleBytesIn': 8,
            'mtxrQueueSimpleBytesOut': 9,
            'mtxrQueueSimplePacketsIn': 10,
            'mtxrQueueSimplePacketsOut': 11,
            'mtxrQueueSimplePCQQueuesIn': 12,
            'mtxrQueueSimplePCQQueuesOut': 13,
            'mtxrQueueSimpleDroppedIn': 14,
            'mtxrQueueSimpleDroppedOut': 15,
        }
        return self._table_entries(
            self.oid_lookup['mtxrQueueSimpleEntry'],
            table_columns,
            column_names
        )

    def mtxrWlApTable(self, column_names):
        table_columns = {
            'mtxrWlApIndex': 1,
            'mtxrWlApTxRate': 2,
            'mtxrWlApRxRate': 3,
            'mtxrWlApSsid': 4,
            'mtxrWlApBssid': 5,
            'mtxrWlApClientCount': 6,
            'mtxrWlApFreq': 7,
            'mtxrWlApBand': 8,
            'mtxrWlApNoiseFloor': 9,
            'mtxrWlApOverallTxCCQ': 10,
            'mtxrWlApAuthClientCount': 11,
        }
        return self._table_entries(
            self.oid_lookup['mtxrWlApEntry'],
            table_columns,
            column_names
        )

    def mtxrWlRtabTable(self, column_names):
        table_columns = {
            'mtxrWlRtabAddr': 1,
            'mtxrWlRtabIface': 2,
            'mtxrWlRtabStrength': 3,
            'mtxrWlRtabTxBytes': 4,
            'mtxrWlRtabRxBytes': 5,
            'mtxrWlRtabTxPackets': 6,
            'mtxrWlRtabRxPackets': 7,
            'mtxrWlRtabTxRate': 8,
            'mtxrWlRtabRxRate': 9,
            'mtxrWlRtabRouterOSVersion': 10,
            'mtxrWlRtabUptime': 11,
            'mtxrWlRtabSignalToNoise': 12,
            'mtxrWlRtabTxStrengthCh0': 13,
            'mtxrWlRtabRxStrengthCh0': 14,
            'mtxrWlRtabTxStrengthCh1': 15,
            'mtxrWlRtabRxStrengthCh1': 16,
            'mtxrWlRtabTxStrengthCh2': 17,
            'mtxrWlRtabRxStrengthCh2': 18,
            'mtxrWlRtabTxStength': 19,
        }
        return self._table_entries(
            self.oid_lookup['mtxrWlRtabEntry'],
            table_columns,
            column_names
        )

    def mtxrWlStatTable(self, column_names):
        """Get Wireless Stats"""
        table_columns = {
            'mtxrWlStatIndex': 1,
            'mtxrWlStatTxRate': 2,
            'mtxrWlStatRxRate': 3,
            'mtxrWlStatStrength': 4,
            'mtxrWlStatSsid': 5,
            'mtxrWlStatBssid': 6,
            'mtxrWlStatFreq': 7,
            'mtxrWlStatBand': 8,
        }
        return self._table_entries(
            self.oid_lookup['mtxrWlStatEntry'],
            table_columns,
            column_names
        )

    def os_version(self):
        """Return the Routerboard OS version"""
        return str(self.get_oid(self.oid_lookup['mtrxLicVersion']))



from distutils.core import setup

setup(
    name = 'yandc',
    packages = ['yandc', 'yandc_base', 'yandc_snmp', 'yandc_ssh', 'yandc_ros'],
    requires = ['paramiko', 'pysnmp', 'pyasn1'],
    version = '0.2c',
    description = 'Yet Another Networked Device Client module',
    author = 'Matt Ryan',
    author_email = 'inetuid@gmail.com',
    url = 'https://github.com/inetuid/yandc',
    download_url = 'https://github.com/inetnet/yandc/tarball/0.2',
    classifiers = [],
)



import yandc_base as base
import yandc_snmp as snmp
from yandc_eos import Client as EOS_Client
from yandc_ios import Client as IOS_Client
from yandc_ros import Client as ROS_Client


class Client(object):
    def __new__(cls, *args, **kwargs):
        assert 'host' in kwargs, 'No host specified'

        grouped_kwargs = base.Utils.group_kwargs('snmp_', **kwargs)

        if 'snmp_' not in grouped_kwargs:
            raise base.ClientError('No SNMP details specified')

        vendor, sys_object_id = snmp.Client(
            kwargs['host'],
            **grouped_kwargs['snmp_']
        ).enterprise()

        if EOS_Client.is_arista(sys_object_id):
            return EOS_Client(*args, **kwargs)
        elif vendor == 'Cisco':
            return IOS_Client(*args, **kwargs)
        elif vendor == 'Cumulus':
            return CL_Client(*args, **kwargs)
        elif ROS_Client.is_mikrotik(sys_object_id):
            return ROS_Client(*args, **kwargs)
        else:
            raise base.ClientError('No driver for vendor')



import re
#
from yandc_ssh import Shell as BaseShell


class Shell(BaseShell):
    def __init__(self, *args, **kwargs):
        self.output_re_sub_regexp = re.compile(r'{}\[(9999B|c)'.format(chr(0x1B)))
        super(Shell, self).__init__(*args, **kwargs)

    def on_output_line(self, *args, **kwargs):
        return re.sub(
            self.output_re_sub_regexp,
            '',
            super(Shell, self).on_output_line(*args, **kwargs)
        ).lstrip('\r')



"""Simple SSH client class"""

import socket
#
import paramiko
#
from .exception import AuthenticationError, ConnectError


class Client(object):
	"""Simple SSH client class"""
	def __del__(self):
		self.disconnect()

	def __enter__(self):
		return self

	def __exit__(self, exception_type, exception_value, traceback):
		self.disconnect()

	def __init__(self, host, **kwargs):
		tcp_port = kwargs.get('port', 22)
		connect_timeout = kwargs.get('timeout', 10)

#		paramiko.common.logging.basicConfig(level=paramiko.common.DEBUG)
		paramiko.common.logging.basicConfig(level=paramiko.common.CRITICAL)

		try:
			sock = socket.create_connection((host, tcp_port), connect_timeout)
			paramiko_transport = paramiko.Transport(sock)
			paramiko_transport.connect()
		except socket.error as error:
			raise ConnectError(
				'Could not connect to {}:{} - [{}]'.format(host, tcp_port, error.message)
			)
		else:
			self.on_connect(paramiko_transport)

		paramiko_transport.set_keepalive(5)

#		auth_types = []

#		try:
#			auth_types = paramiko_transport.auth_none(kwargs['username'])
#		except paramiko.BadAuthenticationType as auth_error:
#			auth_types = auth_error.allowed_types
#		except paramiko.SSHException as ssh_error:
#			raise GeneralError(ssh_error)

		try:
			paramiko_transport.auth_password(kwargs['username'], kwargs['password'])
		except paramiko.AuthenticationException as auth_error:
			raise AuthenticationError(auth_error.message)
		except paramiko.BadAuthenticationType as bad_auth_type:
			raise AuthenticationError(
				'Auth method not supported - [{}]'.format(bad_auth_type.allowed_types)
			)
		else:
			self.paramiko_transport = paramiko_transport

	def __repr__(self):
		return 'ssh.Client @ {} ({})'.format(
			hex(long(id(self)) & long(0xffffffff)),
			repr(self.paramiko_transport)
	)

	def channel(self):
		"""Return a channel object"""
		if 'get_banner' in dir(self.paramiko_transport):
			pass
		return self.paramiko_transport.open_session()

	def disconnect(self):
		if hasattr(self, 'paramiko_transport'):
			if self.is_active():
				self.paramiko_transport.close()
			del self.paramiko_transport

	def exec_command(self, command, *args):
		chan = self.channel()
		chan.set_combine_stderr(True)
		chan.exec_command(command, *args)
		output = chan.makefile('rb')
		exec_output = []
		for output_line in output.readlines():
			exec_output.append(output_line.rstrip('\r\n'))
		chan.shutdown(2)
		chan.close()
		return exec_output

	def is_active(self):
		"""Is the underlying transport active"""
		if hasattr(self, 'paramiko_transport'):
			return self.paramiko_transport.is_active()
		return False

	def on_connect(self, paramiko_transport):
		"""Override to run on connect()"""
		pass

	def sftp_client(self):
		return paramiko.SFTPClient.from_transport(self.paramiko_transport)
